@@ -62,6 +62,7 @@ struct _EMailBrowserPrivate {
 	gulong close_on_reply_response_handler_id;
 
 	guint show_deleted : 1;
+	guint show_junk : 1;
 };
 
 enum {
@@ -75,6 +76,7 @@ enum {
 	PROP_REPLY_STYLE,
 	PROP_MARK_SEEN_ALWAYS,
 	PROP_SHOW_DELETED,
+	PROP_SHOW_JUNK,
 	PROP_UI_MANAGER
 };
 
@@ -276,6 +278,7 @@ mail_browser_message_selected_cb (EMailBrowser *browser,
 	info = camel_folder_get_message_info (folder, uid);
 
 	if (info != NULL) {
+		GSettings *settings;
 		EMailDisplay *display;
 		const gchar *title;
 
@@ -288,8 +291,10 @@ mail_browser_message_selected_cb (EMailBrowser *browser,
 		gtk_window_set_title (GTK_WINDOW (browser), title);
 		gtk_widget_grab_focus (GTK_WIDGET (display));
 
-		camel_message_info_set_flags (
-			info, CAMEL_MESSAGE_SEEN, CAMEL_MESSAGE_SEEN);
+		settings = e_util_ref_settings ("org.gnome.evolution.mail");
+		if (g_settings_get_boolean (settings, "mark-seen"))
+			camel_message_info_set_flags (info, CAMEL_MESSAGE_SEEN, CAMEL_MESSAGE_SEEN);
+		g_clear_object (&settings);
 
 		camel_message_info_unref (info);
 	}
@@ -455,6 +460,12 @@ mail_browser_set_property (GObject *object,
 				g_value_get_boolean (value));
 			return;
 
+		case PROP_SHOW_JUNK:
+			e_mail_browser_set_show_junk (
+				E_MAIL_BROWSER (object),
+				g_value_get_boolean (value));
+			return;
+
 		case PROP_MARK_SEEN_ALWAYS:
 			e_mail_reader_set_mark_seen_always (
 				E_MAIL_READER (object),
@@ -528,6 +539,13 @@ mail_browser_get_property (GObject *object,
 				E_MAIL_BROWSER (object)));
 			return;
 
+		case PROP_SHOW_JUNK:
+			g_value_set_boolean (
+				value,
+				e_mail_browser_get_show_junk (
+				E_MAIL_BROWSER (object)));
+			return;
+
 		case PROP_UI_MANAGER:
 			g_value_set_object (
 				value,
@@ -1057,6 +1075,17 @@ e_mail_browser_class_init (EMailBrowserClass *class)
 			G_PARAM_READWRITE |
 			G_PARAM_STATIC_STRINGS));
 
+	g_object_class_install_property (
+		object_class,
+		PROP_SHOW_JUNK,
+		g_param_spec_boolean (
+			"show-junk",
+			"Show Junk",
+			"Show junk messages",
+			FALSE,
+			G_PARAM_READWRITE |
+			G_PARAM_STATIC_STRINGS));
+
 	g_object_class_install_property (
 		object_class,
 		PROP_UI_MANAGER,
@@ -1214,6 +1243,28 @@ e_mail_browser_set_show_deleted (EMailBrowser *browser,
 	g_object_notify (G_OBJECT (browser), "show-deleted");
 }
 
+gboolean
+e_mail_browser_get_show_junk (EMailBrowser *browser)
+{
+	g_return_val_if_fail (E_IS_MAIL_BROWSER (browser), FALSE);
+
+	return browser->priv->show_junk;
+}
+
+void
+e_mail_browser_set_show_junk (EMailBrowser *browser,
+			      gboolean show_junk)
+{
+	g_return_if_fail (E_IS_MAIL_BROWSER (browser));
+
+	if (browser->priv->show_junk == show_junk)
+		return;
+
+	browser->priv->show_junk = show_junk;
+
+	g_object_notify (G_OBJECT (browser), "show-junk");
+}
+
 GtkUIManager *
 e_mail_browser_get_ui_manager (EMailBrowser *browser)
 {
